package com.android.tcplugins.FileSystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.Scanner;
import java.util.TimeZone;

import org.apache.http.util.EncodingUtils;

import android.content.Context;
import android.webkit.MimeTypeMap;
import android.widget.Toast;

public class PicoServer {
	int port=0;
	ServerSocket serverSocket=null;
	Thread mThread;
	
	public PicoServer (int port1) throws IOException
	{
		port = port1;
		serverSocket = new ServerSocket(port);
		//serverSocket.setSoTimeout(10000);
		if (port==0)  // use system-assigned port!
			port=serverSocket.getLocalPort();
		mThread = new Thread( new Runnable() {
			public void run() {
				while(true) { 
					try {
						final Socket s1=serverSocket.accept();
						new Thread(new Runnable() {
							public void run() {
								Socket s=s1;
								handleHttpRequest(s);
							}
						}).start();
					} catch ( Throwable e ){
						try {
							if (serverSocket.isClosed())
								break;
						} catch ( Throwable e2 ){
							break;
						}
					}
				}
			}
		});
		mThread.setDaemon( true );
		mThread.start();
	}

	private void handleHttpRequest(Socket s) {
		try {
			long rangeStart=0;
			long rangeEnd=-1;
			// only allow local requests
			SocketAddress addr=s.getRemoteSocketAddress();
			boolean allowed=false;
			if (addr instanceof InetSocketAddress) {
				InetAddress addr2=((InetSocketAddress)addr).getAddress();
				if (addr2!=null && addr2.isLoopbackAddress())
					allowed=true;
			}
			if (!allowed) {
				sendErrorReplyAndClose(s,"403 Forbidden");
				return;
			}
			InputStream st = s.getInputStream();
			if (st==null)
				return;
			byte[] buffer=new byte[8192];
			st.read(buffer);  // InputStream directly in Scanner causes a hang on Android 1.6
			Scanner scanner = new Scanner(EncodingUtils.getAsciiString(buffer));
			String line = scanner.nextLine();
			if (!line.startsWith("GET ")) {
				sendErrorReplyAndClose(s,"501 Not Implemented");
				return;
			}
			String path=null;
			if (line.length()>=4) {
				int i=line.indexOf(' ',4);
				if (i>0)
					path=line.substring(4,i);
				i=path.indexOf( '?' );
				if (i>=0)
					path=path.substring(0,i);
				byte[] path2=new byte[line.length()];
				int outidx=0;
				for(i=0;i<path.length();i++) {
					char c = path.charAt(i);
					switch (c) {
						case '+':
							path2[outidx++]=' ';
							break;
						case '%':
							path2[outidx++]=(byte)Integer.parseInt(path.substring(i+1,i+3), 16 );
							i += 2;
							break;
						default:
							path2[outidx++]=(byte)c;
							break;
					}
				}
				path=EncodingUtils.getString(path2, 0, outidx, "UTF-8");				
			}
			if (path==null || path.length()==0 || path.matches(".*\\/\\.{2,}\\/.*")) {
				sendErrorReplyAndClose(s,"400 Bad Request");
				return;
			}
			// now read the headers!
			// we only need the range header
			while (true) {
				try {
					line = scanner.nextLine();
				} catch (Throwable e) {
					break;
				}
				if (line.trim().length()==0)
					break;				
				int i=line.indexOf(':');
				if (i>0 && line.toLowerCase().startsWith("range")) {
					line=line.substring(i+1).toLowerCase();
					i=line.indexOf("bytes=");
					if (i>=0) {
						i+=6;
						line=line.substring(i);
						i=line.indexOf('-');
						if (i>=0) {
							String s2=line.substring(0,i);
							if (s2.length()>0) {
								try { 
									rangeStart=Long.parseLong(s2);
								} catch (Exception e) {
									rangeStart=0;
								}
							}
							s2=line.substring(i+1);
							if (s2.length()>0)
								try { 
									rangeEnd=Long.parseLong(s2);
								} catch (Exception e) {
									rangeEnd=-1;
								}
						}
					}

				}
			}
			sendFileAndClose(s,path,rangeStart,rangeEnd);
			st.close();			
		} catch (Throwable e) {
		}
		
	};
	
	private void sendErrorReplyAndClose(Socket s,String response) {
		try {
			OutputStream st = s.getOutputStream();
			response="HTTP/1.0 "+response+"\r\nContent-Type: text/plain"+
					"\r\n\r\n"+response+"\r\n";
			byte[] buffer=EncodingUtils.getBytes(response,"UTF-8");
			st.write(buffer);
			st.flush();
			st.close();
		} catch (Throwable e) {
		}
		try {
			s.close();
		} catch (Throwable e) {
		}
	}
	
	private void sendFileAndClose(Socket s,String path,long rangeStart,long rangeEnd) {
		FileInputStream fis=null;
		OutputStream st=null;
		try {
			st=s.getOutputStream();
			// here starts the plugin-specific code!
			File f = new File(path);
			try {
				fis = new FileInputStream(f);
				fis.skip(rangeStart);
			} catch (Throwable e) {
				fis=null;
			}
			if (fis==null) {
				sendErrorReplyAndClose(s,"404 Not Found");
				return;
			}
			long length=f.length();
			if (rangeStart>length)
				rangeStart=length;
			if (rangeEnd>length)
				rangeStart=length;
			if (rangeEnd==-1)
				length=length-rangeStart;
			else
				length=rangeEnd-rangeStart;
			String response=createHeader(path,length,f.lastModified(),rangeStart,rangeEnd);
			byte[] buffer=EncodingUtils.getBytes(response,"UTF-8");
			st.write(buffer);
			st.flush();
			// now send the actual data
			int bufferSize=32768;
			buffer = new byte[bufferSize];
			while (length>0) {
				int read=fis.read(buffer,0,length>bufferSize?bufferSize:(int)length);
				if (read<0)
					break;
				st.write(buffer,0,read);
				length-=read;
			}
			st.flush();
		} catch (Throwable e) {
		}
		try {
			if (fis!=null)
				fis.close();
		} catch (Throwable e) {
		}
		try {
			if (st!=null)
				st.close();
		} catch (Throwable e) {
		}
		try {
			s.close();
		} catch (Throwable e) {
		}
	}

	private String createHeader(String path, long length, long timeStamp, long rangeStart,long rangeEnd) {
		SimpleDateFormat dateFormat = new java.text.SimpleDateFormat( "E, d MMM yyyy HH:mm:ss 'GMT'", Locale.US);
		dateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));
		String response;
		if (rangeStart>0 || rangeEnd!=-1)
			response="206 Partial Content";
		else
			response="200 OK";
		// Get MIME type from file name extension, if possible
		String mimetype = null;
		int dot = path.lastIndexOf( '.' );
		if ( dot >= 0 ) {
			MimeTypeMap m=MimeTypeMap.getSingleton();
			mimetype = m.getMimeTypeFromExtension(path.substring(dot+1).toLowerCase());
		}
		if (mimetype==null)
			mimetype="application/octet-stream";
		response="HTTP/1.1 "+response+
		    "\r\nDate: "+dateFormat.format(new Date())+
		    "\r\nServer: Apache/1.0 (Unix)"+
			"\r\nETag: \""+Integer.toHexString((path+timeStamp+""+length).hashCode())+"\""+
			"\r\nAccept-Ranges: bytes"+
			"\r\nContent-Length: "+length+
			"\r\nCache-Control: max-age=604800"+
			"\r\nConnection: close"+
			"\r\nContent-Type: "+mimetype;
		if (timeStamp>=0)
			response+="\r\nLast-Modified: "+dateFormat.format(new Date(timeStamp));
		if (rangeStart>0 || rangeEnd!=-1) {
			if (rangeEnd==-1) {
				response+="\r\nContent-Range: bytes "+rangeStart+"-";
			} else {
				response+="\r\nContent-Range: bytes "+rangeStart+"-"+rangeEnd;
			}
		}		
 		return response+"\r\n\r\n";
	}

	public int getPort() {
		return port;
	}

	public void stop() {
		try
		{
			serverSocket.close();
			mThread.join();
		} catch (Throwable e) {
		}		
	}
}
