package com.android.tcplugins.FileSystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import android.app.Service;
import android.content.Intent;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.net.Uri;
import android.os.Handler;
import android.os.IBinder;
import android.os.RemoteException;
import android.widget.Toast;

import com.android.tcplugins.FileSystem.IPluginFunctions.Stub;
import com.ghisler.tcplugins.sample.R;

public class PluginFunctions extends Stub {
	public static final int  RT_OTHER=0;
	public static final int  RT_USERNAME=1;
	public static final int  RT_PASSWORD=2;
	public static final int  RT_ACCOUNT=3;
	public static final int  RT_USERNAMEFIREWALL=4;
	public static final int  RT_PASSWORDFIREWALL=5;
	public static final int  RT_TARGETDIR=6;
	public static final int  RT_URL=7;
	public static final int  RT_MSGOK=8;
	public static final int  RT_MSGYESNO=9;
	public static final int  RT_MSGOKCANCEL=10;
	public static final int  RT_CHANGEDIR=99;
	public static final int  RT_REFRESH=100;

	public static final int RT_CRYPT_SAVE_PASSWORD=101;
	public static final int RT_CRYPT_LOAD_PASSWORD=102;
	public static final int RT_CRYPT_LOAD_PASSWORD_NO_UI=103; // Load password only if master password has already been entered!
	public static final int RT_CRYPT_COPY_PASSWORD=104;       // Copy encrypted password to new connection name
	public static final int RT_CRYPT_MOVE_PASSWORD=105;       // Move password when renaming a connection
	public static final int RT_CRYPT_DELETE_PASSWORD=106;     // Delete password
	public static final int FS_PUT_COPY_FILE = 1;
	public static final int FS_PUT_MOVE_FILE = 2;
	public static final int FS_GET_COPY_FILE = 4;
	public static final int FS_GET_MOVE_FILE = 8;
	public static final int FS_GET_LOCAL_NAME = 16;
	public static final int FS_EXECUTE = 32;
	public static final int FS_DELETE = 64;
	public static final int FS_MAKEDIR = 128;
	public static final int FS_GET_BITMAP = 256;
	public static final int FS_RENAME = 512;
	public static final int FS_REMOTE_COPY = 1024;
	public static final int FS_STREAM = 2048;
	public static final int FS_HIDE_TRANSFER_MODE = 4096;
	public static final int FS_REMOTE_SOURCE = 8192;
	public static final int FS_REMOTE_TARGET = 16384;
	public static final int FS_GET_ENCODED_REMOTE_NAME = 32768;
	public static final int FS_REQUIRES_CONTENT_URLS = 0x20000; // The plugin cannot request access to the file system, we must send content: urls
	public static final int FS_SUPPORTS_CONTENT_URLS = 0x40000;
	public static final int FS_DELETE_YOURSELF_UPLOAD_MOVE = 0x80000; // The host should delete the local file by itself after a move returning 0

	public static final int FS_STATUS_INFO_NEEDED = 2048;
	public static final int FS_STATUS_LANGUAGE_CHANGED = 4096;
	public static final int FS_STATUS_THEME_CHANGED = 4097;

	private static final int FS_COPYFLAGS_MOVE = 1;
	private static final int FS_COPYFLAGS_OVERWRITE = 2;
	private static final int FS_COPYFLAGS_RESUME = 4;
	private static final int FS_COPYFLAGS_EXISTS_SAMECASE = 8;
	private static final int FS_COPYFLAGS_EXISTS_DIFFERENTCASE = 16;
	private static final int FS_COPYFLAGS_REQUEST_DELETE = 32;

	// flags for tLogProc
	public static final int MSGTYPE_CONNECT=1;
	public static final int MSGTYPE_DISCONNECT=2;
	public static final int MSGTYPE_DETAILS=3;
	public static final int MSGTYPE_TRANSFERCOMPLETE=4;
	public static final int MSGTYPE_CONNECTCOMPLETE=5;
	public static final int MSGTYPE_IMPORTANTERROR=6;
	public static final int MSGTYPE_OPERATIONCOMPLETE=7;
	public static final int MSGTYPE_REFRESH=8;

// ids for FsGetFile
	private static final int FS_FILE_OK=0;
	private static final int FS_FILE_EXISTS=1;
	private static final int FS_FILE_NOTFOUND=2;
	private static final int FS_FILE_READERROR=3;
	private static final int FS_FILE_WRITEERROR=4;
	private static final int FS_FILE_USERABORT=5;
	private static final int FS_FILE_NOTSUPPORTED=6;
	private static final int FS_FILE_EXISTSRESUMEALLOWED=7;

	private static final int FS_EXEC_OK=0;
	private static final int FS_EXEC_ERROR=1;
	private static final int FS_EXEC_YOURSELF=-1;
	private static final int FS_EXEC_SYMLINK=-2;

	private Service service=null;
	private boolean aborted=false;
	private IRemoteProgressCallback mProgressCallback=null;
	private IRemoteDialogCallback mDialogCallback=null;
	private PicoServer picoServer=null;
	
	public PluginFunctions(Service service2) {
		service=service2;
	}
	
	@Override
	public boolean deleteFile(String fileName) throws RemoteException {
		/*final String s="Delete: "+fileName;
    	mMessageHandler.post(new Runnable() {
            public void run() {
        		Toast.makeText(service, s, Toast.LENGTH_SHORT).show();
            };
    	});*/
		boolean ret=false;
		try {
			File f=new File(fileName);
			if (f.isFile())
				ret=f.delete();
		} catch (Exception e) {
			ret=false;
		}
		return ret;
	}

	@Override
	public boolean removeDir(String dirName) throws RemoteException {
		try {
			File f=new File(dirName);
			if (f.isDirectory())
				return f.delete();
		} catch (Exception e) {
		}
		return false;
	}

	@Override
	public boolean makeDir(String dirName) throws RemoteException {
		try {
			File f=new File(dirName);
			return f.mkdir();
		} catch (Exception e) {
		}
		return false;
	}
	
	@Override
	public Bitmap getBitmap(String fileName) throws RemoteException {
		// TODO Auto-generated method stub
		return null;
	}

    private Handler mMessageHandler = new Handler();
	private String listCacheName="";
	private ArrayList<ArrayList<PluginItem>> listCache=null;
	public static final int MAX_ITEMS=5000;

	@Override
	public List<PluginItem> getDirectoryList(String dirName)
			throws RemoteException {
		try {
			int pathPos=dirName.indexOf('/');
			if (pathPos>0) {
				int cacheIndex=Integer.parseInt(dirName.substring(0, pathPos));
				dirName=dirName.substring(pathPos);
				if (dirName.equals(listCacheName) && cacheIndex>=0 && cacheIndex<listCache.size()) {
					List<PluginItem> res=listCache.get(cacheIndex);
					if (cacheIndex==listCache.size()-1) {
						listCacheName="";
						listCache.clear();
						listCache=null;
					}
					return res;
				} else
					return null;
			}			
			ArrayList<ArrayList<PluginItem>> allLists=new ArrayList<ArrayList<PluginItem>>(); 
			File f=new File(dirName);
			if (f.isDirectory()) {
				File[] list=f.listFiles();
				ArrayList<PluginItem> res;
				int remainItems=0;
				if (list.length>MAX_ITEMS) {
					res=new ArrayList<PluginItem>(MAX_ITEMS+1);  // 1 more to signal continued list!
					remainItems=list.length-MAX_ITEMS;
				} else
					res=new ArrayList<PluginItem>(list.length);
				int i=0;
				int i2=0;
				int i3=0;
				for (File f2:list) {
					if (aborted)
						return null;
					PluginItem obj=new PluginItem();
					obj.name=f2.getName();
					obj.description=null;
					obj.directory=f2.isDirectory();
					obj.lastModified=f2.lastModified();
					obj.length=f2.length();
					obj.iconFlags=0;
					res.add(obj);
					i++;
					i2++;
					i3++;
					if (i2>=500) {
						mProgressCallback.logProc(MSGTYPE_DETAILS, i3+"/"+list.length);
						i2=0;
					}
					if (i>=MAX_ITEMS && remainItems!=0) {
						obj=new PluginItem();
						obj.name="\t";   // tells TC that there is more data!
						res.add(obj);
						i=0;
						allLists.add(res);
						if (remainItems<=MAX_ITEMS) {
							res=new ArrayList<PluginItem>(remainItems);
							remainItems=0;
						} else {
							res=new ArrayList<PluginItem>(MAX_ITEMS);
							remainItems-=MAX_ITEMS;
						}
					}
				}
				if (allLists.isEmpty())
					return res;
				else {
					allLists.add(res);   // last, incomplete part
					listCacheName=dirName;
					listCache=allLists;
					return listCache.get(0);
				}
			}
		} catch (Exception e) {
		}
		return null;
	}

	@Override
	public String getLocalFileName(String fullSrc) throws RemoteException {
		return null;
	}

	@Override
	public int getSupportedFunctions() throws RemoteException {
		return FS_DELETE | FS_PUT_COPY_FILE | FS_PUT_MOVE_FILE | FS_GET_COPY_FILE | FS_GET_MOVE_FILE |
			FS_GET_LOCAL_NAME | FS_EXECUTE | FS_MAKEDIR | FS_RENAME | FS_REMOTE_COPY | FS_STREAM |
			FS_HIDE_TRANSFER_MODE | FS_REMOTE_SOURCE | FS_REMOTE_TARGET;
	}

	@Override
	public int execute(String[] fullPath, String verb)
			throws RemoteException {
		if (verb.equalsIgnoreCase("startserver")) {
			if (picoServer==null) {
				try {
					picoServer=new PicoServer(0);
				} catch (Throwable e) {
					return FS_EXEC_ERROR;
				}
			}
			fullPath[0]="http://127.0.0.1:"+picoServer.getPort();
			return FS_EXEC_OK;
		} else if (verb.equalsIgnoreCase("stopserver")) {
			if (picoServer!=null) {
				try {
					picoServer.stop();
				} catch (Exception e) {
					return FS_EXEC_ERROR;
				}
				picoServer=null;
			}
			return FS_EXEC_OK;
		} else if (verb.equalsIgnoreCase("open")) {
			return FS_EXEC_YOURSELF;
		} else if (verb.compareTo("properties")==0) {
			if (fullPath[0].equals("/"))
				showAboutBox(); 
		}
		return FS_EXEC_ERROR; 
	}
	
    private void showAboutBox() {
        PackageInfo pInfo;
        String versionInfo="";
		try {
			pInfo = service.getPackageManager().getPackageInfo(service.getPackageName(), PackageManager.GET_META_DATA);
			versionInfo = pInfo.versionName;
		} catch (Exception e) {
			versionInfo = "";
		}
		String appName=service.getString(R.string.app_name);
		String about=appName+" "+versionInfo+
		"\nCopyright ©\n2011-2017 C. Ghisler,\nGhisler Software GmbH\nhttp://www.ghisler.com";
    	try {
			mDialogCallback.requestProc(RT_MSGOK, appName, about);
		} catch (Exception e) {
		}
	}

	@Override
	public int putFile(String fullSrc, String fullTrg, int copyFlags)
			throws RemoteException {
		File f=new File(fullSrc);
		if (!f.isFile())
			return FS_FILE_NOTFOUND;
		if ((copyFlags & FS_COPYFLAGS_OVERWRITE)==0)
			// The sender tells us that the remote file already exists!
			// Our file system is case sensitive, so we need to react to
			// both case-sensitive and insensitive matches
			if ((copyFlags & (FS_COPYFLAGS_EXISTS_SAMECASE | FS_COPYFLAGS_EXISTS_DIFFERENTCASE))!=0) {
				return FS_FILE_EXISTS;
			}
			
		return copyFile(fullSrc,fullTrg,copyFlags,f.length(),f.lastModified());
	}
	
	@Override
	public int getFile(String fullSrc, String[] fullTrg, int copyFlags,
			long size, long lastWriteTime) throws RemoteException {
		return copyFile(fullSrc,fullTrg[0],copyFlags,size,lastWriteTime);
	}

	private int copyFile(String fullSrc, String fullTrg, int copyFlags,
			long size, long lastWriteTime) {
		boolean bFailIfExists=(copyFlags & FS_COPYFLAGS_OVERWRITE)==0;
		boolean bMove=(copyFlags & FS_COPYFLAGS_MOVE)!=0;

		boolean ok=true;
		int err=0;

		File gf=new File(fullTrg);
		if (gf.exists()) {
			File ff=new File(fullSrc);  // Make sure we don't copy the file onto itself
			String s1;
			try {
				s1 = ff.getCanonicalPath();
			} catch (IOException e) {
				s1=fullSrc;
			}
			String s2;
			try {
				s2 = gf.getCanonicalPath();
			} catch (IOException e) {
				s2 = fullTrg;
			}
			if (s1.equals(s2)) {
				return FS_FILE_WRITEERROR;
			} else
				if (!bFailIfExists)
					gf.delete();
				else
					return FS_FILE_EXISTS;
		}
		if (!bMove) {
			int bufferSize=65536;
			byte[] buffer = new byte[bufferSize];
			FileInputStream f=null;
			try {
				f = new FileInputStream(fullSrc);
			} catch (FileNotFoundException e1) {
				err=FS_FILE_NOTFOUND;
				ok=false;
			}
			FileOutputStream g=null;
			try {
				g = new FileOutputStream(fullTrg);
			} catch (FileNotFoundException e1) {
				err=FS_FILE_WRITEERROR;
				ok=false;
			}
			long done=0;
			long oldpercent=0;
			int bytesread=0;
			if (ok)	do {
				try {
					bytesread=f.read(buffer);
				} catch (Exception e) {
					err=FS_FILE_READERROR;
					ok=false;
					break;
				}
				if (aborted)
					break;
				if (bytesread>0) {
					try {
						g.write(buffer, 0, bytesread);
					} catch(Exception e) {
						err=FS_FILE_WRITEERROR;
						ok=false;
						break;
					}
					done+=bytesread;
					if (size>0) {
						long percent=done*100/size;
						if (percent!=oldpercent) {
							if (mProgressCallback!=null)
								try {
									mProgressCallback.progressPercent(percent>100?100:(int)percent);
								} catch (Exception e) {
								}
							oldpercent=percent;
						}
					}
					
				}
			} while (bytesread>0);
			try {
				g.close();
			} catch (Exception e) {
				ok=false;
				err=FS_FILE_WRITEERROR;
			}
			if (aborted || err!=0) {
				try {
					gf.delete();
				} catch (Exception e) {
				}
			} else {
				try {
					if (gf!=null && lastWriteTime!=0 && lastWriteTime!=-1)
						gf.setLastModified(lastWriteTime);
				} catch (Exception e) {
				}
			}
			try {
				f.close();
			} catch (Exception e) {
				ok=false;
				err=FS_FILE_READERROR;
			}
		} else {
			File ff=new File(fullSrc);
			if (ff.isFile() || gf.isDirectory()) {
				try {
					if (gf.isFile())
						if (bFailIfExists)
							err=FS_FILE_EXISTS;
						else
							gf.delete();
					if (err!=FS_FILE_EXISTS) {
						ok=ff.renameTo(gf);
						if (!ok)
							err=FS_FILE_READERROR;
					}
				} catch (Exception e) {
					ok=false;
					err=FS_FILE_READERROR;
				}
			}
		}
		if (!ok) {
			if (aborted)
				err=FS_FILE_USERABORT;
			else {
				if (err==FS_FILE_READERROR) {
					File ff=new File(fullSrc);
					if (!ff.isFile())
						err=FS_FILE_NOTFOUND;
				}
			}
		}
		return err;
	}

	@Override
	public void registerCallbacks(IRemoteProgressCallback cb,IRemoteDialogCallback cb2)
			throws RemoteException {
		mProgressCallback=cb;
		mDialogCallback=cb2;
	}

	@Override
	public int renMovFile(String fullSrc, String fullTrg, boolean move,
			boolean overwrite, long size, long lastWriteTime) throws RemoteException {
		int copyFlags=0;
		if (move)
			copyFlags |= FS_COPYFLAGS_MOVE;
		if (overwrite)
			copyFlags |= FS_COPYFLAGS_OVERWRITE;
		return copyFile(fullSrc,fullTrg,copyFlags,size,lastWriteTime);
	}

	@Override
	public void setAbortFlag(String dirName, boolean abort)
			throws RemoteException {
		aborted=abort;		
	}

	@Override
	public String disconnect(String rootName) throws RemoteException {
		// unused
		return null;
	}

	@Override
	public void statusInfo(String dirName, int infoStartEnd, int infoOperation)
			throws RemoteException {
		// unused;
	}

	class CopyCallback extends IRemoteCopyCallback.Stub {
		String fileName=null;
		InputStream f=null;
		
		public CopyCallback(String fullSrc) throws FileNotFoundException,IOException {
			fileName=fullSrc;
		}

		@Override
		public boolean open(long resumeAt) throws RemoteException {
			try {
				f=new FileInputStream(fileName);
				if (resumeAt==0)
					return f!=null;
				else {
					if (f.skip(resumeAt)==resumeAt)
						return true;
					else {
						close();
						return false;
					}
				}
			} catch (Throwable e) {
				return false;
			}
		}
		
		@Override
		public int read(byte[] buffer, int size) throws RemoteException {
			try {
				return f.read(buffer, 0, size);
			} catch (Throwable e) {
				return -1;
			}
		}

		@Override
		public void close() throws RemoteException {
			try {
				if (f!=null)
					f.close();
			} catch (Throwable e) {
			}
			f=null;
		}
	}
	
	@Override
	public IRemoteCopyCallback getFileCallback(String fullSrc) throws RemoteException {
		CopyCallback mCopyCallback = null;
		try {
			mCopyCallback = new CopyCallback(fullSrc);
		} catch (Throwable e) {
		}
		return mCopyCallback;
	}
	
	@Override
	public int putFileFromCallback(IRemoteCopyCallback in, String fullTrg, int copyFlags, long size, long lastWriteTime) throws RemoteException {
		boolean OverWrite=(copyFlags & FS_COPYFLAGS_OVERWRITE)!=0;
		boolean Resume=(copyFlags & FS_COPYFLAGS_RESUME)!=0;
		long resumeAt=0;

		if ((copyFlags & (FS_COPYFLAGS_EXISTS_SAMECASE | FS_COPYFLAGS_EXISTS_DIFFERENTCASE))!=0 &&
			!(OverWrite | Resume))
			return FS_FILE_EXISTSRESUMEALLOWED;
		if (aborted)
			return FS_FILE_USERABORT;

		if (Resume) {
			try {
				File f=new File(fullTrg);
				resumeAt=f.length();
			} catch (Throwable e) {
				return FS_FILE_WRITEERROR;
			}
		}
		
		if (size<resumeAt || !in.open(resumeAt))
			return FS_FILE_READERROR;
		
		long percent = 0;
		RandomFileOutputStream out = null;
  	  	try {
			out = new RandomFileOutputStream(fullTrg);
			if (out == null) {
				throw new IOException("could not get Output Stream for '"+fullTrg+"'");
			}
			if (resumeAt>0)
				out.setFilePointer(resumeAt);  // trows exception if failed
			else
				out.setFileSize(0);
			
			byte[] buf = new byte[32768];
			
			long len = size-resumeAt;
			int cnt=0;
			do {
				cnt = in.read(buf, buf.length);
				out.write(buf, 0, cnt);
				if (cnt <= 0 || aborted) {
					throw new RemoteException();
				}
				len -= cnt;
				if (size>0) {
					percent=(size-len)*100/size;
					mProgressCallback.progressPercent(percent>100?100:(int)percent);
				}
			} while (cnt>0);
			out.flush();
			in.close();
			out.close();
			return 0;
		} catch (Exception e) {
			if (out != null) {
				try {
					out.close();
				}
				catch (Exception ignore) {} 
			}
			if (in != null) {
				try {
					in.close();
				}
				catch (Exception ignore) {} 
			}
			throw new RemoteException();
		}
	}

	@Override
	public String getModifiedLocalName(String fullRemote, String localName) throws RemoteException {
		// only used for Google Drive so far:
		// original name may be "Document", converted to
		// PDF it will become "Document.pdf".
		return null;
	}
	
}
